/* Common functions for NBIAS */

// Random number generator from 'Numeric recipies'
#define IA 16807
#define IM 2147483647
#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS DBL_EPSILON /*1.2e-15*/
#define RNMX (1.0-EPS)

// Random number generator

static long _idum_;
static int _ran1_not_initialized_=1;

double ran1(double range)
{
	int j;
	long k;
	static long iy=0;
	static long iv[NTAB];
	double temp;
   
    if(_ran1_not_initialized_) {printf("\nERROR: RAN1() was not initialized");exit(1);}

	if (_idum_ <= 0 || !iy) {
		if (-(_idum_) < 1) _idum_=1;
		else _idum_ = -(_idum_);
		for (j=NTAB+7;j>=0;j--) {
			k=(_idum_)/IQ;
			_idum_=IA*(_idum_-k*IQ)-IR*k;
			if (_idum_ < 0) _idum_ += IM;
			if (j < NTAB) iv[j] = _idum_;
		}
		iy=iv[0];
	}
	k=(_idum_)/IQ;
	_idum_=IA*(_idum_-k*IQ)-IR*k;

if (_idum_ < 0) _idum_ += IM;
	j=iy/NDIV;
	iy=iv[j];
	iv[j] = _idum_;
	if ((temp=AM*iy) > RNMX) return (RNMX*range);
	else return(temp*range);
}

void init_ran1(long seed)
{
int i;

if(seed == 0)
{
    srand((unsigned)time( NULL ));
    for(i = 0;i < 100;i++) _idum_ = rand();
    /*printf("SEED=%ld\n",_idum_);*/
}
else _idum_ = seed;

if(_idum_ > 0) _idum_ = -_idum_;

if(_idum_ == 0) _idum_ = -123456789;

_ran1_not_initialized_ = 0;
for(i = 0;i < 100;i++) ran1(1000.);
}


#undef IA
#undef IM
#undef AM
#undef IQ
#undef IR
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX


// Convert amino acid character into a number
int ConvToInt(char c)
{
int i;

for(i = 0;i < 20;i++) if(c == aa[i]) return(i);
return(20);
}

// Read 'double' from a string read from a text file
double ReadDblS(FILE *in)
{
char _str_[200];

fgets(_str_,200,in);
return(atof(_str_));
}

void form_out_l(FILE *fil,char *s,int l,int break_len)
{
int i,counter;
char c;

for(i = 0,counter = 0;i < l;i++)
{
 counter++;
 putc(s[i],fil);
 if(counter == break_len) {fputc('\n',fil);counter = 0;}
}
if(counter) fputc('\n',fil);
}
//NOTE: min separation default = 3

/*
  bool is_bound
  Description: Determines if we've hit either the sequence bound, or if we've hit a bad position.
*/
bool is_bound(int bound, int seq_length, char *mark)
{
//Are we beyond the beginning of the sequence?
    if (bound < 1) {
	return true;
    }
//Are we beyond the end of the sequence?
    if (bound > seq_length) {
	return true;
    }
//Are we at a bad position?
    if (mark[bound-1] == 0){
	return true;
    }
//Otherwise, we're OK, not a bound.
    return false;

}

/* bool past_threshold
 Description: Used with ExtendSegments, this function determines whether we're below or above the threshold, depending on whether we're in high or low mode.
*/
bool past_threshold(double avg, double threshold, char high_low) {
//Are we looking at the high or low threshold?
    if (high_low == 'h'){
	if (avg < threshold) {
	    return true;
	}
    }
    if (high_low == 'l'){
	if (avg > threshold){
	    return true;
	}
    }
//Otherwise, we haven't passed the threshold.
    return false;
}

bool compare(const struct FRAGMENT &s1,const struct FRAGMENT &s2)
{
if(s1.start < s2.start) return(true);
return(false);
}

// Probability of finding one position with particular properties given a set of amino acids ('name' is a '\0' terminated string)
// and average frequencies 'frq'
double ComputeResidueP(char *name,double *frq)
{
int i,j;
double p=0.;

for(i = 0;i < strlen(name);i++)
{
    j = ConvToInt(name[i]);
    if(j < 20) p += frq[j];
}

return(p);
}

// Probability of finding one position with particular properties given a set of nucleotides ('name' is a '\0' terminated string)
// and average frequencies 'frq'
double ComputeResiduePNT(char *name,double *frq)
{
int i,j;
double p=0.;

for(i = 0;i < strlen(name);i++)
{
    j = FindLetter(name[i],4,nt);
    if(j != -1) p += frq[j];
}

return(p);
}

// Computes the number of positions occupied by specific amino acids in a subsequence of 'seq' given by 'start' and 'end'
// 'seq' is a zero-indexed string 
int ComputeNSuccesses(char *name,char *seq,int start,int end)
{
int i,j,n;

for(i = start,n = 0;i <= end;i++)
for(j = 0;j < strlen(name);j++)
{
    if(seq[i] == name[j]) {n++;break;}
}

return(n);
}

int ComputeN(char *seq,int start,int end)
{
int i,n;

for(i = start,n = 0;i <= end;i++) if(seq[i] == '1') n++;

return(n);
}

/*--------------- Integration of normal distribution using Simpson's rule -----------------*/

double NormalDensity(double x)
{
return(0.398942280401432677939946059934382*exp(-0.5*x*x));
}

#define FUNC(x) ((*func)(x))
   
double trapzd(double (*func)(double), double a, double b, int n)
{
	double x,tnm,sum,del;
	static double s;
	int it,j;
   
	if (n == 1) 
    {
		return (s=0.5*(b-a)*(FUNC(a)+FUNC(b)));
	} 
    else 
    {
		for (it=1,j=1;j<n-1;j++) it <<= 1;
		tnm=it;
		del=(b-a)/tnm;
		x=a+0.5*del;
		for (sum=0.0,j=1;j<=it;j++,x+=del) sum += FUNC(x);
		s=0.5*(s+(b-a)*sum/tnm);
		return s;
	}
}

#undef FUNC

#define EPS DBL_EPSILON
#define JMAX 20
   
double qsimp(double (*func)(double), double a, double b)
{
	double trapzd(double (*func)(double), double a, double b, int n);
	int j;
	double s,st,ost,os;
   
	ost = os = -1.0e30;
	for (j=1;j<=JMAX;j++) {
		st=trapzd(func,a,b,j);
		s=(4.0*st-ost)/3.0;
		if (j > 5)
			if (fabs(s-os) < EPS*fabs(os) ||
				(s == 0.0 && os == 0.0)) return s;
		os=s;
		ost=st;
	}
	/*printf("\nToo many steps in routine qsimp()\n");*/
	return s;
}

#undef EPS
#undef JMAX

/*-------------------------------------------------*/

////////////////////////////////////////////////////////////
// routines to compute P(successes) in N bernoulli trials //
////////////////////////////////////////////////////////////

// Normal approximation to Binomial distribution
double ApproxBinomial_P_GreaterThan_K(int k,int n,double p,int &flag)
{
double m,s,npq,approx_p;

m = (double)n*p;
npq = (double)n*p*(1.0-p);
if(npq < 5) {fprintf(stderr,"\nNormal approximation to Binomial distribution is not valid\n");flag = 0;}
else flag = 1;
s = sqrt(npq);

approx_p = qsimp(NormalDensity,((double)k-0.5-m)/s,((double)k-0.5-m)/s+10.);
return(approx_p);
}

double ApproxBinomial_P_LessThan_K(int k,int n,double p,int &flag)
{
double m,s,npq,approx_p;

m = (double)n*p;
npq = (double)n*p*(1.0-p);
if(npq < 5) {fprintf(stderr,"\nNormal approximation to Binomial distribution is not valid\n");flag = 0;}
else flag = 1;
s = sqrt(npq);

approx_p = qsimp(NormalDensity,((double)k+0.5-m)/s-10.,((double)k+0.5-m)/s);
return(approx_p);
}

double factorial(int n)
{
int i;
double val=1.;

if(n == 0) return(1.);
for(i = 1;i <= n;i++) val *= (double)i;
return(val);
}

double n_Choose_k(int k,int n)
{
int i;
double val=1.;

for(i = n-k+1;i <= n;i++) val *= (double)i;
return(val/factorial(k));
}

/* Binomial probability for P(x=k), n - total # of trials, p - prob.of success */
double BinomialProb(int k,int n,double p)
{
return( n_Choose_k(k,n)*pow(p,k)*pow(1.-p,n-k) );
}

/* Recursion to compute binomial probability */
double BPR(int k,int n,double p)
{
double val;

if(k == 0) return(BinomialProb(0,n,p));
val = BPR(k-1,n,p);
//if(val <= DBL_MIN) val = 0.;
return((p/(1.-p))*val*((double)(n-k+1)/(double)k));
}

/* Cumulative binomial probability for less than k successes in n trials */
double Binomial_P_LessThan_K(int k,int n,double p)
{
int i;
double val=0.,pr;

for(i = 0;i < k;i++) 
{
    pr = BPR(i,n,p);
    if(pr >= DBL_MIN) val += pr;
}

if(val >= 1.-DBL_EPSILON) return(1.);
else return(val);
}

double Binomial_P_GreaterThan_K(int k,int n,double p)
{
int i;
double val=0.,pr;

for(i = n;i >= k;i--) 
{
    pr = BPR(i,n,p);
    if(pr >= DBL_MIN) val += pr;
}

if(val >= DBL_MIN) return(val);
else return(0.);
}


/////////////////////////////////////////////////////////////////////////////////


/* Checks if a character already exists in array 's' */
int AAExists(char c,char *s,int n)
{
int i;

if(n == 0) return(0);
for(i = 0;i < n;i++) if (c == s[i]) return(1);
return(0);
}

/* Recodes amino acid into '0' or '1' */
char AA2binary(char c,char *s,int n)
{
int i;

if(n == 0) return('0');
for(i = 0;i < n;i++) if (c == s[i]) return('1');
return('0');
}

/* Copies only valid a.a. characters skipping repeated characters */
void CopyAA(char *source,char *dest)
{
int i,j;

for(i = 0,j = 0;j < strlen(source);j++) 
{
    if(ConvToInt(toupper(source[j])) != 20 && AAExists(toupper(source[j]),dest,i) == 0) {dest[i] = toupper(source[j]);i++;}
}
dest[i] = '\0';
}

/* Copies only valid nucleotide characters skipping repeated characters */
void CopyNT(char *source,char *dest)
{
int i,j;

for(i = 0,j = 0;j < strlen(source);j++) 
{
    if(FindLetter(toupper(source[j]),4,nt) != -1 && AAExists(toupper(source[j]),dest,i) == 0) {dest[i] = toupper(source[j]);i++;}
}
dest[i] = '\0';
}

// Cumulative probability of observing value of >= k estimated using array 'prob'
double CumulativeProb(int k,int l,double *prob)
{
int i;
double p=0.;

for(i = k;i <= l;i++) p += prob[i];
return(p);
}

// Exact probability of observing longest run of k successes in n Bernoulli trials with the prob. of success p (Bateman)
double ExactRunProb(int k,int n,double p)
{
int i,j;
double prob=0.;

for(j = 1;j <= n/k;j++)
{
    prob += pow(-1.,(double)(j+1))*(p+((double)(n-j*k+1)*(1.-p)/(double)j))*n_Choose_k(j-1,n-j*k)*pow(p,(double)(j*k))*pow((1.-p),(double)(j-1));
}

return(prob);
}

// Approximate probability of observing longest run of k successes in n Bernoulli trials with the prob. of success p 
double ApproxRunProb(int k,int n,double p)
{
double Q2,Q3,q,prob;

q = 1.-p;
Q2 = 1.-pow(p,k)*(1.+(double)k*q);
Q3 = 1.-pow(p,k)*(1.+2.*(double)k*q)+0.5*pow(p,2*k)*(2.*(double)k*q+(double)k*(double)(k-1)*q*q);
prob = 1.-Q2*pow(Q3/Q2,n/k-2);
return(prob);
}

// The probability of observing at least one window of size 'w' that has 'k' or more successes in a sequence of length 'n'
// 'p' - probability of success
double P_ge_K_in_W_N(int k,int w,int n,double p)
{
int wn;
double wp;

// The probability of observing a particular window with at least k successes
wp = Binomial_P_GreaterThan_K(k,w,p);
// Correction for the total of n/w disjoint windows
if(n%w == 0) wn = n/w;
else wn = n/w + 1;
return( 1. - pow(1.-wp,wn) );
}

/////////////////////////// Functions for Equation 4.3 from Glaz et al
// begin
double Fb(int r,int s,double p)
{
int i;
double v = 0.;

if(r < 0) return(0);

for(i = 0;i <= r;i++)
{
    v += BPR(i,s,p);
}

return(v);
}

double A1(int k,int m,double p)
{
    return( 2.*BPR(k,m,p)*Fb(k-1,m,p)*((double)(k-1)*Fb(k-2,m,p)-(double)m*p*Fb(k-3,m-1,p)) );
}

double A2(int k,int m,double p)
{
double b;
    
    b = BPR(k,m,p);
    return( 0.5*b*b*((double)(k-1)*(double)(k-2)*Fb(k-3,m,p)-2.0*(double)(k-2)*(double)m*p*Fb(k-4,m-1,p)+(double)m*(double)(m-1)*p*p*Fb(k-5,m-2,p)) );
}

double A3(int k,int m,double p)
{
int r;
double a = 0.,f;

for(r = 2;r <= k-1;r++)
{   
    f = Fb(r-1,m,p);
    a += BPR(2*k-r,m,p)*f*f;
}

return(a);
}

double A4(int k,int m,double p)
{
int r;
double a = 0.;

for(r = 2;r <= k-1;r++)
{   
    a += BPR(2*k-r,m,p)*BPR(r,m,p)*((double)(r-1)*Fb(r-2,m,p)-(double)m*p*Fb(r-3,m-1,p));
}

return(a);
}

double Q2(int k,int m,double p)
{
double f;

    f = Fb(k-1,m,p);
    return( f*f-(double)(k-1)*BPR(k,m,p)*Fb(k-2,m,p)+(double)m*p*BPR(k,m,p)*Fb(k-3,m-1,p) );
}

double Q3(int k,int m,double p)
{
double f;
    
    f = Fb(k-1,m,p);
    return( f*f*f-A1(k,m,p)+A2(k,m,p)+A3(k,m,p)-A4(k,m,p) );
}

// The probability of observing scan statistic Sw >= 'k' in a sequence of length 'n'
// 'p' - probability of success
double P_scan_ge_K_in_W_N(int k,int w,int n,double p)
{
int wn;
double ret;

if(n%w == 0) wn = n/w;
else wn = n/w + 1;

ret = 1. - Q2(k,w,p)*pow(Q3(k,w,p)/Q2(k,w,p),wn-2);

if(ret < 0.) return(0); // If p-value is too close to zero, there may be too many accumulated floating-point errors  
else return(ret);
}
// End 
///////////////////////////////////////////////////

// Merges discrete runs of 'special' residues
int MergeSegmentsDiscrete(char *seq, std::vector<FRAGMENT>& segs,int max_sep //Max allowed separation between segments
			             )
{
    int i,j,density;
	int max_length = 0;
	vector<FRAGMENT>::iterator iter;
	vector<FRAGMENT>::iterator iter2;

//For every segment, see if any of the rest overlap.
for (iter = segs.begin(); iter != segs.end();iter++)
{
		for (iter2 = iter+1; iter2 != segs.end();iter2++)
		{

			//If the end of i is greater to or equal to the start of j - max_sep, merge the segs.
			if ((*iter).end >= (*iter2).start - max_sep)
			{

			//Only change i's end if it's less than j's.
			if((*iter2).end > (*iter).end) {
				(*iter).end=(*iter2).end;
				(*iter).l = (*iter).end - (*iter).start + 1;
				}

			segs.erase(iter2);
			iter2 = iter;
			}

		}

}

//Now, calculate the density of special residudes in each segment and max_length
	
	for (iter = segs.begin(); iter != segs.end(); ++iter)
	{
		  (*iter).l = (*iter).end - (*iter).start + 1;
		  density = ComputeN(seq,(*iter).start-1,(*iter).end-1);

		  (*iter).n = density;
	
		//See if this segment's the longest one.
		if ((*iter).l > max_length){
			max_length = (*iter).l;
			}
		}

		return max_length;
}

// Puts all runs of characters from 'letters' found in 'seq' into 'segs'
// Returns length of the longest run
int FindRuns(char *seq,int l,std::vector<FRAGMENT>& segs)
{
int i,j,start,end,f_size,max_length;
char first;
struct FRAGMENT p;

first = seq[0];
start = 0;
max_length = 0;

for(j = 1;j < l;j++)
{
        if(seq[j] != first) // end of the run
        {
            end = j-1;
            f_size = end-start+1;

			// Save only runs of 1's
			if(first == '1')
			{
				p.start = start+1;
				p.end = end+1;
				p.l = f_size;
				p.n = ComputeN(seq,start,end);
				segs.push_back(p);
				if(max_length < f_size) max_length = f_size;
			}

            first = seq[j];
            start = j;
        }   
		if(j == l-1 && seq[j] == first && first == '1') // end of the sequence
		{
				end = j;
				p.start = start+1;
				p.end = end+1;
				p.l = end - start + 1;
				p.n = ComputeN(seq,start,end);
				segs.push_back(p);
				if(max_length < p.l) max_length = p.l;
		}
}

return(max_length);
}

// Slides scanning window of size 'w' along a binary 0,1 sequence of length 'l',
// saves all windows with p-value < 'p' into 'dat'
// Returns largest number of '1's in the window
int SlideWindowP(char *seq,int l,int w,double p,std::vector<FRAGMENT>& dat)
{
int i,j,counter,max_density=0;
double pvalue;
struct FRAGMENT f;

for(i = 0;i < l-w+1;i++)
{
	for(counter = 0,j = i;j < i+w;j++) if(seq[j] == '1') counter++;
	if(max_density < counter) max_density = counter;
    pvalue = P_scan_ge_K_in_W_N(counter,w,l,P_special);
	if(pvalue < p)
	{
		f.start = i+1;
		f.end = i+w;
		f.l = f.end - f.start + 1;
		f.n = counter;
		f.prob = pvalue;
		dat.push_back(f);
	}
}

return(max_density);
}

// Slides scanning window of size 'w' along a binary 0,1 sequence of length 'l',
// Returns largest number of '1's in the window
int SlideWindow(char *seq,int l,int w)
{
int i,j,counter,max_density=0;
struct FRAGMENT f;

for(i = 0;i < l-w+1;i++)
{
	for(counter = 0,j = i;j < i+w;j++) if(seq[j] == '1') counter++;
	if(max_density < counter) max_density = counter;
}

return(max_density);
}

// Compute nucleotide frequencies from sequence 'seq'
void ComputeNTfreq(char *seq,int l,double *frq)
{
int i,n,k;

for(i = 0;i < 4;i++) frq[i] = 0.;

for(i = 0,n = 0;i < l;i++)
{
   if((k=FindLetter(seq[i],4,nt)) != -1) {frq[k]++;n++;continue;}
}

if(n == 0) {fprintf(stderr,"\nNo valid nucleotides were found\n");exit(1);}

for(i = 0;i < 4;i++) frq[i] /= (double)n;
}

void Conv2Upper(char *seq,int l)
{
int i;

for(i = 0;i < l;i++) seq[i] = toupper(seq[i]);
}
int EstimateLinkageDistance(double p)
{
int d;

d = (int)round(2./(3.*p),"%.0lf");
if(d == 0) d = 1;

return(d);
}
